/*
############################################################################## 
# BibItem.java
#
# Stellt ein BibItem Element dar.
# Hierbei werden die im BibItem vorhandenen Elemente (address, url,
# ...) in einer Hashtable aufbewahrt, wobei der Hashschlssel der
# Elementname ist und der Hashwert der Elementwert.
#
# Beispiel BibItem:
# <bibitem key="ID9641745026230" type="InProceedings">
#	<address>Chalkidiki, Griechenland</address>
# 	<url>http://www.ibr.cs.tu-bs.de/ harbaum/projekte/hpcs97.ps.gz</url>
#       <year>1997</year>
#	<author>M. Zitterbart and T. Harbaum and D. Meier</author>
# 	<month>jun</month>
#	<booktitle>IEEE Workshop HPCS 1997</booktitle>
#	<title>Efficient Routing Table Lookup for IPv6</title>
# </bibitem>
# Jonas Pohlandt 2002
##############################################################################
*/
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServlet;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.w3c.dom.*;

import java.util.*;
import java.io.*;
import java.lang.Integer;

import java.util.regex.*;

public class BibItem {

    // ## private Variablen ##     

    private Hashtable elements   = new Hashtable();

    // ErrorHandler der beim parsen von XML verwendet wird.
    MySaxErrorHandler mySaxEH;

    // ## Konstruktoren ##

    // Leerer Konstruktor    
    public BibItem() {}
    
    // Konstruiert ein BibItem aus einem HttpServletRequest. Die
    // Biblist wird bentigt, um den Schlssel (das "key" Element)
    // auf Einmaligkeit zu prfen.
    //
    // Parameter     req: HTTP Request aus dem das BibItem erstellt werden soll
    //	         biblist: BibList, welche zur berprfung der einmaligkeit
    //                    des Schlssels verwedet wird.
    public BibItem(HttpServletRequest req, BibList biblist) {
	
        Enumeration parameternames;
        String      parameter;
        String      value;
        Date        date = new Date();

	// Alle Parameternamen (title, year, ...) des Requests in eine
	// Enumeration schreiben
        parameternames = sortParameters(req.getParameterNames());

	// Jedes BibItem bentigt einen eindeutigen Schlssel. Dieser
	// wird im "key" Element gespeichert.  Die generateKey()
	// Methode generiert einen gltigen Schlssel, falls Mglich
	// aus Authorname und Jahr.
	// Sonst ID + getTimeInMillis();
        elements.put("key", generateKey(req, biblist));
        
	// Alle Parameternamen und -werte aus dem Request in die
	// entsprechende Hashtable schreiben.
        while (parameternames.hasMoreElements()) {
            parameter = (String) (parameternames.nextElement());
            value     = req.getParameter(parameter);

            // Leere Eingabefelder und Metadaten ignorieren
            if (value.equals("") || isMetaData(parameter)) {
                continue;
            }
	    
            elements.put(parameter, value);
        }
    }
    
    // Konstruiert ein BibItem aus einem HttpServletRequest. Der
    // Schlssel wird hier mit bergeben und mu damit nicht generiert
    // werden. Ansonsten ist dieser Konstruktor mit dem vorhergehenden
    // identisch.
    //
    // Parameter     req: HTTP Request aus dem das BibItem erstellt werden soll
    //		uniqueid: Schlsselattribut des neuen BibItem
    public BibItem(HttpServletRequest req, String uniqueid) {

        Enumeration parameternames;
        String      parameter;
        String      value;
        Date        date = new Date();

        parameternames = sortParameters(req.getParameterNames());
        elements.put("key", uniqueid);

        while (parameternames.hasMoreElements()) {
            parameter = (String) (parameternames.nextElement());
            value     = req.getParameter(parameter);

            if (value.equals("") || isMetaData(parameter)) {
                continue;
            }

            elements.put(parameter, value);
        }
    }
    
    // Konstruiert eine BibItem aus einem Node (der wiederum selbst
    // ein BibItem darstellen muss)
    //
    // Parameter 	node: Node aus dem das BibItem erstellt werden soll  
    public BibItem(HttpServlet s, Node node) {

        // Alle im BibItem enthaltenen Elemente in eine NodeList schreiben
        NodeList childNodesList = node.getChildNodes();

	elements.put("key", ((Element)node).getAttribute("key"));
	elements.put("type", ((Element)node).getAttribute("type"));
	// s.log("entering " + ((Element)node).getAttribute("type") + " " +
	//      ((Element)node).getAttribute("key"));

        // Fr alle Nodes (= Elemente) in der NodeList
        for (int j = 0; j < childNodesList.getLength(); j++) {
	    //j-ter KindKnoten (= Element) des Bibitem
            Node childNode = childNodesList.item(j);
            Element element = (Element) node;
	    
            if (childNode.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE) {
		// Nur Elemente berprfen
                
                // Das in childNode enthaltene Element, also
                // z.B. <year>1999</year> enthlt wiederum mehrere
                // ChildNodes. Zum Beispiel den TextNode welcher 1999
                // enthlt.
                NodeList elementChildsList = childNode.getChildNodes();

                for (int k = 0; k < elementChildsList.getLength(); k++) {
		    // Alle Kinder des Elementes prfen
                    Node cnode = elementChildsList.item(k);

		    // s.log("  entering " + childNode.getNodeName() + ": " + cnode.getNodeValue());
                    if (cnode.getNodeType() == org.w3c.dom.Node.TEXT_NODE) {
			// Nur der text interessiert
                        elements.put(childNode.getNodeName(),
                                     cnode.getNodeValue());
                    }
                }
            }
        }
    }
    
    // Methoden fr den direkten Zugriff auf die Hashtable
    public String getElement(String element) {
    	String elementvalue = (String) elements.get(element);
    	return elementvalue;  	
    }
    
    public void setElement(String element, String value) {
    	elements.put(element, value);
    }
    
    // Eine Hilfsmethode welche bei der Generierung eines Schlssels
    // verwendet wird.  authors enthlt den Text des <author> Elements
    // (z.B. M. Zitterbart, A.Pinger) Es wird davon ausgegangen wird,
    // dass Authoren immer die Form
    // "<Grossbuchstabe><Punkt><Leerzeichen><Grossbuchstabe>
    // <Kleinbuchstabe(mindestens einer)> haben.
    //
    // Parameter: Der die Authoren enthaltende String
    //
    // Rckgabewert: Anzahl der Authoren als int
    private int countAuthors(String authors) {
        int count = 0;
        String[] s;
        Pattern pattern = Pattern.compile("[A-Z]\\.\\s[A-Z][a-z]+");
        Matcher matcher = pattern.matcher(authors);
        if (matcher.find()) {
            count++;
        } else {
            return count;
        }
        while (matcher.find(matcher.end())) {
            count++;
        }
        return count;	
    }
    
    // Diese Methode generiert einen eindeutigen Schlssel fr ein
    // BibItem. Die Form des Schlssels hngt dabei von der Anzahl der
    // Authoren ab:
    // 0 Authoren: key -> ID<Zeit seit 1970 in Millisekunden>
    // 1 Author: key -> <Erste 3 Buchstaben des Nachnamens>
    //                  <Letzte 2 Stellen des Verff. Jahres>
    // 2+ Authoren: key -> <Anfangsbuchstaben der Nachnamen der Authoren>
    //                     <Letzte 2 Stellen des Verff. Jahres>
    // Ist kein Verff. Jahr angegeben wird stattdessen XX verwendet.
    // Sollte ein Schlssel bereits Vorhanden sein, wird ein a
    // Angehngt, falls dieser auch schon voranden ist ein b
    // usw. (z.B. Fis98, Fis98a, Fis98b, ...)
    //
    // Parameter:   req: Der die Felder <author> und <year> enthaltende
    //                   HTTP Request
    //		biblist: BibList, welche zur berprfung der einmaligkeit
    //                   des Schlssels verwedet wird.
    //
    // Rckgabewert: der erzeigte key als String
    private String generateKey(HttpServletRequest req, BibList biblist) {
        String uniqueid = "";
        String author;
        String year;
        Date        date = new Date();
        char suffix;
        int i = 97; //  new char(97) entspricht a
        boolean nouse;

        Pattern pattern =
	    Pattern.compile("[A-Z]\\.\\s([A-Z])([a-z]{1,2})[a-z]*");
        author = req.getParameter("author");
        if (author.equals("")) { 
            uniqueid = "ID" + java.lang.Long.toString(date.getTime());
            return uniqueid;
        }
        year = req.getParameter("year");
        if (year.equals("")) { 
            year = "XX";
        }
        
        Matcher matcher = pattern.matcher(author);
                
        if (countAuthors(author) == 0) {
            uniqueid = "ID" + java.lang.Long.toString(date.getTime());
        } else if (countAuthors(author) == 1) {
            nouse = matcher.find();
            uniqueid = matcher.group(1) + matcher.group(2);
        } else {
            nouse = matcher.find();
            uniqueid = matcher.group(1);
            while (matcher.find(matcher.end())) {
            	uniqueid += matcher.group(1);
            }
        }
        uniqueid += year.substring(2);
        
        if (biblist.getBibItem(uniqueid) == null) {
            return uniqueid;
        } else {
            suffix = (char) i;	
            while (!(biblist.getBibItem(uniqueid + suffix) == null)) {
                suffix = (char) ++i;
            }
        }
        uniqueid += suffix;
        return uniqueid;
    }
    
    // Diese Methode erweitert das BibItem um Elemente die den
    // Verleihstatus angeben.  Diese zustzliche Information wird nie
    // in der XML Dateiabgespeichert sondern wird lediglich bei der
    // Anzeige verwendet.
    //
    // Parameter: Das LendItem, welches die Verleihinformation fuer
    // dieses BibItem enthaelt

    public void addLendInfo(LendItem lenditem) {
        if (!(lenditem == null)) {
	    if (!(lenditem.getElement("personref") == null)) {
	        setElement("lend-personref", lenditem.getElement("personref"));
	    }
	    // if (!(lenditem.getElement("lentbymatno") == null)) {
            //     setElement("lentbymatno", lenditem.getElement("lentbymatno"));
            // }
            if (!(lenditem.getElement("date") == null)) {
                setElement("lend-date", lenditem.getElement("date"));
            }
        }	
    }

    // Diese Methode findet das zum BibItem zugehrige LendItem (falls
    // es denn eines gibt) Dies wird erreicht, in dem das Lendfile
    // nach einem LendItem durchsucht wird, das eines mit dem key
    // Element des BibItems bereinstimmendes key Element hat.
    //
    // Parameter: Das Fileobjekt, welches auf die Datei zeigt, welche
    // die LendList enthlt in der das zugehrige LendItem ist.
    //
    // Rckgabewert: Das zum BibItem gehrige LendItem falls
    // vorhanden, sonst null
    public LendItem getLendItem (File lendfile) {
        
        LendItem lenditem = null;
        
        // Aus dem File wird ein DOM Document erzeigt
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(true);
        
        try { 
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setErrorHandler(mySaxEH);
            Document lenddoc = builder.parse(lendfile);
            
            String uniqueid = getElement("key");
            
            // Alle LendItems auf bereinstimmung prfen
            NodeList lendlist = lenddoc.getElementsByTagName("lenditem");
            for (int j = 0; j < lendlist.getLength(); j++) {
		//j-ter LendItem Node
                lenditem = new LendItem(lendlist.item(j));
                if (lenditem.getElement("key").equals(uniqueid)) {
                    break;
                } else {
                    lenditem = null;
                }
            }
	
	// Mgliche Exceptions abfangen
        } catch (SAXException sxe) {
            Exception x = sxe;
            if (sxe.getException() != null) {
                x = sxe.getException();
            }
            x.printStackTrace();
        } catch (ParserConfigurationException pce) {
            pce.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
        return lenditem;    	
    }
    
    // Diese Methode prft ob das BibItem ausgeliehen ist. 
    //
    // Parameter: Die LendList in der das zugehrige LendItem stehen knnte.
    //
    // Rchgabewert: true wenn ausgeliehen, sonnst false
    public boolean isLent(LendList lendlist) {
        boolean lent = false; 
        LendItem lenditem;
        Enumeration e;
                
        String uniqueid = getElement("key");
        e = lendlist.keys();
        while (e.hasMoreElements()) {               
        	
            // exisitiert ein LendItem mit dem key des BibItems ist es
            // ausgeliehen
            if (e.nextElement().equals(uniqueid)) {
                lent = true;
                break;
            }
        }
        return lent;    	
    }


    // Diese Methode gibt einen Node zurck der dem BibItem entspricht.
    // Rckgabewert: der erzeugte Node
    public Node toNode() {

        Node                   node;
        Element                element = null;
        Document               document;
        Enumeration            e;
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

        factory.setValidating(true);

        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setErrorHandler(mySaxEH);

            document = builder.newDocument();
            element  = (Element) document.createElement("bibitem");

	    element.setAttribute("key", (String) elements.get("key"));
	    element.setAttribute("type", (String) elements.get("type"));

            e = elements.keys();

            while (e.hasMoreElements()) {
                String  key   = (String) e.nextElement();

		if (!key.equals("key") && !key.equals("type")) {
		    Element child = (Element) document.createElement(key);

		    child.appendChild((Text)document.createTextNode(
                                                 (String) elements.get(key)));
		    element.appendChild(child);
		}
            }
        
        // Mgliche Exceptions abfangen
        } catch (ParserConfigurationException pce) {
            pce.printStackTrace();
        }

        node = (Node) element;

        return node;
    }


    // Diese Methode prft ob das BibItem auf die im Request angegebenen
    // Parameter passt.
    //
    // Parameter: Der zu prfende HTTP Request
    //
    // Rckgabewert: true falls alle im Request vorhandenen Parameter
    // mit dem BibItem bereinstimmen, sonst false.
    public boolean matches(HttpServlet s, HttpServletRequest req) {

        boolean     match         = true;
        boolean     casesensitive = false;
        Enumeration parameternames;
        String      parameter;
        String      boolvalue = new String("");
        String      value;
        String[]    values;
        String      nodevalue;
        String      reqvalue;
        Pattern pattern;
        Matcher matcher;

        // req.getParameter("csbool") ist nicht null falls im Web
        // Formular "Klein-/Groschreibung unterscheiden" aktiviert
        // war
        if (!(req.getParameter("csbool") == null)) {
            casesensitive = true;
        }

        parameternames = req.getParameterNames();

	// Alle Parameternamen (title, year, ...) des Requests in eine
	// Enumeration schreiben
        while (parameternames.hasMoreElements()) {

            if (!match) {
                break;
            }

            parameter = (String) (parameternames.nextElement());
            value     = req.getParameter(parameter);
            value     = value.trim();
		
	    // Leere Eingabefelder und Metadaten ignorieren
            if (value.equals("") || isMetaData(parameter)) {
                continue;
            }

	    // s.log("checking " + parameter);

            boolvalue = req.getParameter(parameter + "bool");
            
	    nodevalue = (String) elements.get(parameter);
		
	    // Ist das Element nicht vorhanden schlgt die Suche
	    // bei diesem BibItem fehl.
	    if (nodevalue == null) {
		match = false;
		break;
	    }
            
            match = false;
            
            // Daten mssen gesondert betrachtet werden
            // Das einzige Element im Datumsformat ist <lenddate> 
            if (parameter.matches(".*date")) {
            	// Das im Web Formular angegebene Datum
            	GregorianCalendar parameterdate = parseDate(value); 
            	// Das im BibItem stehende Datum
            	GregorianCalendar nodedate = parseDate(nodevalue); 
           
            	if (boolvalue.equals("=")) {
            	    match = nodevalue.matches(value);
            	    break;
            	}
            	
            	if (boolvalue.equals("<=")) {
            	    match = ((nodevalue.matches(value)) ||
			     (nodedate.before(parameterdate)));
            	    break;
            	}
            	
            	if (boolvalue.equals(">=")) {
            	    match = ((nodevalue.matches(value)) ||
			     (nodedate.after(parameterdate)));
            	    break;
            	}
            	
            	if (boolvalue.equals("<")) {
            	    match = nodedate.before(parameterdate);
            	    break;
            	}
            	
            	if (boolvalue.equals(">")) {
            	    match = nodedate.after(parameterdate);
            	    break;
            	}
            	
            }
	    
            int nodeint;
            int reqint;
            
            // values ist ein String Array. splitValues trennt einen
            // String nach Leerzeichen auf und gibt ein Array zurck,
            // welches die einzelnen Strings enthlt.
            // Bsp: String s = new String("a b c");
            // String[] sa = splitValues(s); -> String[1] ist jetzt "b"
	    values    = splitValues(value);
	    
	    // Jetzt werden alle "Wrter" des Feldes, welches gerade
	    // geprft wird durchgegeangen.  Wurde z.B. in der
	    // Suchmaske bei Author "Marx Engels Lenin" einegeben,
	    // luft die Schleife drei mal durch.
            for (int v = 0; v < values.length; v++) {
                
                // reqvalue enthlt nacheinander zuerst Marx dann
                // Engels und schliesslich Lenin, falls die Schleife
                // nicht schon vorher durch ein break verlassen wurde,
                // z.B. wenn bei verUNDeter Suche ein Wort nicht
                // gefunden wurde.
                reqvalue = values[v];

		// Zuerst werden Integer Werte geprft. Ist ein zum
		// Parameter zugehriger Boolparameter <, >, <=, >=
		// oder = wird ein Integerwert angenommen
                if (boolvalue != null) try {
                    if (boolvalue.equals("=")) {
                        nodeint = (new Integer(nodevalue)).intValue();
                        reqint  = (new Integer(reqvalue)).intValue();

                        if (nodeint == reqint) {
                            match = true;
                            break;
                        } else {
                            match = false;
                            break;
                        }
                    }

                    if (boolvalue.equals("<=")) {
                        nodeint = (new Integer(nodevalue)).intValue();
                        reqint  = (new Integer(reqvalue)).intValue();

                        if (nodeint <= reqint) {
                            match = true;
                            break;
                        } else {
                            match = false;
                            break;
                        }
                    }

                    if (boolvalue.equals(">=")) {
                        nodeint = (new Integer(nodevalue)).intValue();
                        reqint  = (new Integer(reqvalue)).intValue();

                        if (nodeint >= reqint) {
                            match = true;
                            break;
                        } else {
                            match = false;
                            break;
                        }
                    }

                    if (boolvalue.equals("<")) {
                        nodeint = (new Integer(nodevalue)).intValue();
                        reqint  = (new Integer(reqvalue)).intValue();

                        if (nodeint < reqint) {
                            match = true;
                            break;
                        } else {
                            match = false;
                            break;
                        }
                    }

                    if (boolvalue.equals(">")) {
                        nodeint = (new Integer(nodevalue)).intValue();
                        reqint  = (new Integer(reqvalue)).intValue();

                        if (nodeint > reqint) {
                            match = true;
                            break;
                        } else {
                            match = false;
                            break;
                        }
                    }
                
                // Falls kein Integerwert im Eingabefeld war muss die
                // Exception gefangen werden
                } catch (NumberFormatException nfe) {
                    System.out.println(
                        "\n** BibItem: Fehler beim umwandeln von String" +
			" zu Integer (Kein Integer in Element " + parameter +
			"." + " BibItem mit key: " + getElement("key") +
			". Parameter wird ignoriert");

                    match = true;
                    break;
                }
                
                // Hatte es sich nicht um ein einen Integerwert
                // enthaltendes Feld gehandelt, wird jetzt per
                // regulaerem Ausdruck versucht Text zu "matchen".
                
                // Pattern erzeugen, abhngig davon, ob Gro- und
                // Kleinschreibung unterschieden werden soll oder
                // nicht.
                if (casesensitive) {
                    pattern = Pattern.compile(values[v]);    
                } else {
                    pattern = Pattern.compile(values[v], 
					      Pattern.CASE_INSENSITIVE);
                }
                
                // Der matcher gibt mit .find() zurck ob mindestens
                // ein Vorkommen des Patterns in nodevalue vorhanden
                // ist.
                matcher = pattern.matcher(nodevalue);
		match = matcher.find();
		
		// Bei verODERter Suche reicht ein Einzeltreffer fr
		// einen Gesamttreffer
                if ((boolvalue != null) && boolvalue.equals("OR")) {
                    if (match) {
                        break;
                    }
                }
			
		// Bei verUNDeter Suche reicht ein einzelner
		// Fehlschlag fr einen Gesamtfehlschlag.
                if ((boolvalue == null) || (!boolvalue.equals("OR"))) {
                    if (!match) {
                        break;
                    }
                }
            }
        }

        return match;
    }

    // Diese Methode liest aus einem String im Format TT MM JJJJ die
    // Zahlenwerte und erzeugt damit ein Objekt vom Typ
    // GregorianCalendar. Ob der String das richtige Format hat muss
    // nicht geprft werden, da dies bereits im Servlet bei der
    // Eingabe eines neuen BibItems geschieht.
    //
    // Parameter: Der zu parsende String der Form TT MM JJJJ
    //
    // Rckgabewert: der erzeugte GregorianCalendar
    private GregorianCalendar parseDate(String s) {
    	GregorianCalendar gc = new GregorianCalendar();
    	int day = Integer.parseInt(s.substring(0, 2));
    	int month = Integer.parseInt(s.substring(3, 5));
    	int year = Integer.parseInt(s.substring(6));
    	gc.set(year, month, day);
    	return gc;
    }
    
    // Diese Methode trennt einen String bei den Leerzeichen auf und
    // gibt ein Array zurck.
    // Bsp: String s = new String("a b c");
    // String[] sa = splitValues(s); -> String[1] ist jetzt "b"
    //
    // Parameter: Der zu splittende String
    //
    // Rckgabewert: Das Array mit den einzelnen Strings
    private String[] splitValues(String s) {

        int    blankCount = 0;
        int    i          = 0;
        String string     = s;
         
        // Zuerst wird gezhlt, wieviele Leerzeichen im String
        // enthalten sind.  Diese Information wird zum Anlegen eines
        // gengend grossen Array bentigt, welches spter die
        // einzelnen Strings enthalten soll.
        
        // Solange noch Leerzeichen im String enthalten sind:
        while (string.indexOf(" ") != -1) {
            blankCount++;
             
            // Erstes Leerzeichen und alle chars links davon entfernen
            string = string.substring(string.indexOf(" ") + 1);
        }

        // Neues Array mit entsprechender Gre anlegen
        String[] values = new String[blankCount + 1];

        // Array fllen
        while (s.indexOf(" ") != -1) {
            values[i] = s.substring(0, s.indexOf(" "));
            s         = s.substring(s.indexOf(" ") + 1);

            i++;
        }

        // Letzter String wird nicht mehr von der while Schleife
        // erfasst, deshalb manuell:
        values[i] = s;

        return values;
    }

    // Diese Methode sortiert eine Enumeration alphanumerisch
    //
    // Parameter: Die zu sortierende Enumaration
    //
    // Rckgabewert: die sortierte Enumeration
    private Enumeration sortParameters(Enumeration e) {

        Vector v = new Vector(0);

        while (e.hasMoreElements()) {
            v.addElement(e.nextElement());
        }

        java.util.Collections.sort(v);

        return v.elements();
    }

    // Diese Methode erkennt Metadaten.
    // Rckgabewert: true falls MetaDaten, sonst false    
    private boolean isMetaData(String s) {
        boolean isMetaData;
        isMetaData = s.matches(".*(database|bool|requester|button|uniqueid)");
        return isMetaData;	
    }


}
